const g = "id", p = "parentId", P = "children";
class o {
  /**
   * Constructor of class
   * @param config - to configure class, if configuration option is omitted, default one is used
   */
  constructor(e) {
    this.idProp = (e == null ? void 0 : e.idProp) || g, this.parentIdProp = (e == null ? void 0 : e.parentIdProp) || p, this.childrenProp = (e == null ? void 0 : e.childrenProp) || P;
  }
  /**
   * Convert list to tree like structure
   * @param list list of objects, objects need to have id (as you configured, or 'id' by default) and parentId property (as you configured, or 'parentId' by default)
   * @param rootParentId - id of root parent nodes (if not specified, root nodes are nodes with parentId of null)
   * @returns tree structure
   */
  list2Tree(e, t = null) {
    return e.filter((r) => r[this.parentIdProp] === t).map((r) => ({
      ...r,
      [this.childrenProp]: this.list2Tree(e, r[this.idProp])
    }));
  }
  /**
   * Convert tree like structure to list
   * @param tree - tree of objects, objects need to have children (as you configured, or 'children' by default) and parentId property (as you configured, or 'parentId' by default)
   * @returns list
   */
  tree2List(e) {
    return this._tree2List(o.deepCopy(e));
  }
  /**
   * Convert tree like structure to list (helper method)
   * @param tree tree of objects, objects need to have children (as you configured, or 'children' by default) and parentId property (as you configured, or 'parentId' by default)
   * @param parentId - id of parent node
   * @private
   * @returns list
   */
  _tree2List(e, t = null) {
    return o.deepCopy(e).reduce((r, i) => {
      const { [this.childrenProp]: s, ...n } = i;
      return [
        ...r,
        { ...n, [this.parentIdProp]: t },
        ...s.length ? this._tree2List(s, n[this.idProp]) : []
      ];
    }, []);
  }
  /**
   * Method to get node in tree structure by given id
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns found node
   */
  get(e, t) {
    return this.find(e, (r) => r[this.idProp] === t);
  }
  /**
   * Method to find node in tree structure by given callback function
   * @param tree - tree structure to search in
   * @param fn - callback function to find node
   * @returns found node
   * @example
   * ```ts
   * utils.find(tree, item => item.id === myId);
   * ```
   */
  find(e, t) {
    const r = e.find((i) => t(i));
    return r || e.reduce(
      (i, s) => i || this.find(s[this.childrenProp] || [], t),
      null
    );
  }
  /**
   * Method to find all nodes in tree structure by given callback function
   * @param tree - tree structure to search in
   * @param fn - callback function to find all nodes
   * @returns all found nodes
   * @example
   * ```ts
   * utils.filter(tree, item => item.id === myId);
   * ```
   */
  filter(e, t) {
    const r = e.filter((i) => t(i));
    return e.reduce((i, s) => [...i, ...s[this.childrenProp].length ? this.filter(s[this.childrenProp], t) : []], r);
  }
  /**
   * Method to delete node in tree by given id (mutable operation!)
   * @param tree - tree structure for node deleting
   * @param id - identifier of node to delete
   */
  delete(e, t) {
    const r = e.findIndex((i) => i[this.idProp] == t);
    return r != -1 ? e.splice(r, 1)[0] : e.reduce((i, s) => i || this.delete(s[this.childrenProp], t), null);
  }
  /**
   * Method to add new node to tree, node will be added as last child (mutable operation!)
   * @param tree - tree structure for node adding
   * @param parentId - identifier of parent node, null if new node should be on root level
   * @param childData - data of new node
   * @param anotherChildData - data of new additional nodes
   */
  add(e, t, r, ...i) {
    this._add("push", e, t, r, ...i);
  }
  /**
   * Method to add new node to tree, node will be added as last child (mutable operation!)
   * @param tree - tree structure for node adding
   * @param parentId - identifier of parent node, null if new node should be on root level
   * @param childData - data of new node
   * @param anotherChildData - data of new additional nodes
   */
  addUnshift(e, t, r, ...i) {
    this._add("unshift", e, t, r, ...i);
  }
  /**
   * Method to add new node to tree at the end or at the beginning (helper method)
   * @param operation - how item should be added into tree
   * @param tree - tree structure for node adding
   * @param parentId - identifier of parent node, null if new node should be on root level
   * @param childData - data of new node
   * @param anotherChildData - data of new additional nodes
   */
  _add(e, t, r, i, ...s) {
    if (r == null) {
      t[e](i, ...s);
      return;
    }
    const n = t.findIndex((h) => h[this.idProp] == r);
    if (n != -1) {
      t[n][this.childrenProp][e](
        { [this.childrenProp]: [], ...i },
        ...s.map((h) => ({ [this.childrenProp]: [], ...h }))
      );
      return;
    }
    t.forEach((h) => this.add(h[this.childrenProp], r, i, ...s));
  }
  /**
   * Method to update node by id with given data in tree (mutable operation!)
   * @param tree - tree structure for node editing
   * @param id - identifier of node to be updated
   * @param data - new data of node (you should also pass children if you want to keep it)
   */
  edit(e, t, r) {
    const i = e.findIndex((s) => s[this.idProp] == t);
    if (i != -1) {
      e[i] = { [this.idProp]: e[i][this.idProp], [this.childrenProp]: [], ...r };
      return;
    }
    e.forEach((s) => this.edit(s[this.childrenProp], t, r));
  }
  /**
   * Method to get descendant nodes of given node in tree structure
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns all found children nodes
   */
  getDescendants(e, t) {
    const r = this.get(e, t);
    return r ? this._getDescendants(r) : [];
  }
  /**
   * Helper method to recursively get all descendant nodes of given node in tree structure
   * @param node - we want to get all of its children
   * @private
   * @returns all found children nodes
   */
  _getDescendants(e) {
    return [
      ...e[this.childrenProp],
      ...e[this.childrenProp].reduce((t, r) => [...t, ...this._getDescendants(r)], [])
    ];
  }
  /**
   * Method to get ancestors of given node in tree structure
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns all found parent nodes
   */
  getAncestors(e, t) {
    const r = [];
    let i = this.getParent(e, t);
    for (; i; )
      r.push(i), i = this.getParent(e, i[this.idProp]);
    return r.reverse();
  }
  /**
   * Method to get nodes that are part of path from root
   * Alias for {@link TreeUtils.getAncestors | getAncestors} method
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns all nodes that are part of path (ordered from root)
   */
  getPathNodes(e, t) {
    return this.getAncestors(e, t);
  }
  /**
   * Method to get parent of given node in tree structure
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns found parent node, otherwise null
   */
  getParent(e, t) {
    return this._getParent(e, t);
  }
  /**
   * Method to get parent of given node in tree structure
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @param parent - parent node, if we found something (for recursion only)
   * @returns found parent node, otherwise null
   */
  _getParent(e, t, r = null) {
    return e.find((s) => s[this.idProp] === t) ? r : e.reduce(
      (s, n) => s || this._getParent(n[this.childrenProp] || [], t, n),
      null
    );
  }
  /**
   * Method to get children of given node in tree structure
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns children nodes of node
   */
  getChildren(e, t) {
    var r;
    return ((r = this.get(e, t)) == null ? void 0 : r[this.childrenProp]) || [];
  }
  /**
   * Method to get neighbours (neighbour is parent or child) of given node in tree structure
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns neighbours of node
   */
  getNeighbours(e, t) {
    return [this.getParent(e, t), ...this.getChildren(e, t)].filter((r) => r);
  }
  /**
   * Method to get siblings of given node in tree structure
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns siblings of node
   */
  getSiblings(e, t) {
    var r;
    return (((r = this.getParent(e, t)) == null ? void 0 : r[this.childrenProp]) || []).filter((i) => i[this.idProp] !== t);
  }
  /**
   * Method to get leafs of subtree from given node
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns leafs of nodes
   */
  getLeafs(e, t) {
    return this.filter(this.getSubTree(e, t), (r) => !r[this.childrenProp].length);
  }
  /**
   * Method to get subtree from given node (children of node)
   * Alias for {@link TreeUtils.getChildren | getChildren} method
   * @param tree - tree structure to search in
   * @param id - identifier of node
   * @returns subtree
   */
  getSubTree(e, t) {
    return this.getChildren(e, t);
  }
  /**
   * Method to get size of subtree (number of nodes in the subtree)
   * @param tree - tree structure
   * @param id - identifier of node
   * @returns size of subtree
   */
  getSize(e, t) {
    return this.tree2List(this.getSubTree(e, t)).length + 1;
  }
  /**
   * Method to get breath of subtree (the number of leaves in subtree)
   * @param tree - tree structure
   * @param id - identifier of node
   * @returns breath of subtree
   */
  getBreath(e, t) {
    return this.getLeafs(e, t).length;
  }
  /**
   * Method to get depth of node (the depth of a node is the length of the path to its root i.e., its root path)
   * @param tree - tree structure
   * @param id - identifier of node
   * @returns depth of node
   */
  getDepth(e, t) {
    return this.getPathNodes(e, t).length;
  }
  /**
   * Method to get level of node (the level of a node is the number of edges along the unique path between it and the root node)
   * @param tree - tree structure
   * @param id - identifier of node
   * @returns level of node
   */
  getLevel(e, t) {
    return this.getDepth(e, t) + 1;
  }
  /**
   * Method to get degree of node (for a given node, its number of children. A leaf, by definition, has degree zero)
   * @param tree - tree structure
   * @param id - identifier of node
   * @returns degree of node
   */
  getDegree(e, t) {
    return this.getChildren(e, t).length;
  }
  /**
   * Method to get degree of tree (the degree of a tree is the maximum degree of a node in the tree)
   * @param tree - tree structure
   * @returns degree of tree
   */
  getTreeDegree(e) {
    return e.reduce((t, r) => Math.max(t, r[this.childrenProp].length, this.getTreeDegree(r[this.childrenProp])), 0);
  }
  /**
   * Method to get nodes in tree at specific level
   * @param tree - tree structure to search in
   * @param level - desired level
   * @returns all nodes, that are on specific level
   */
  getNodesAtLevel(e, t) {
    return this._getNodesAtLevel(e, t);
  }
  /**
   * Helper method to get nodes in tree at specific level recursively
   * @param tree - tree structure to search in
   * @param level - desired level
   * @param actualLevel - actual level, that is searched
   * @private
   * @returns all nodes, that are on specific level
   */
  _getNodesAtLevel(e, t, r = 0) {
    return e.reduce((i, s) => [
      ...i,
      ...t === r ? [s] : [],
      ...r < t ? this._getNodesAtLevel(s[this.childrenProp], t, r + 1) : []
    ], []);
  }
  /**
   * Method to get width on level in tree (the number of nodes in a level)
   * @param tree - tree structure
   * @param level - desired level
   * @returns width on desired level
   */
  getWidth(e, t) {
    return this.getNodesAtLevel(e, t).length;
  }
  /**
   * Method to get height of node (the height of a node is the length of the longest downward path to a leaf from that node)
   * @param tree - tree structure
   * @param id - identifier of node
   * @returns height of node
   */
  getHeight(e, t) {
    return this.getHeightNode(this.getSubTree(e, t));
  }
  /**
   * Helper method to get height of node from children recursively
   * @param tree - tree structure
   * @param height - actual computed height
   * @private
   * @returns height of node
   */
  getHeightNode(e, t = 0) {
    return e.reduce((r, i) => Math.max(r, this.getHeightNode(i[this.childrenProp], t + 1)), t);
  }
  /**
   * Method to get distance between 2 nodes
   * @param tree - tree structure
   * @param id1 - identifier of first node
   * @param id2 - identifier of second node
   * @returns distance between 2 nodes, returns -1, if there is no connection between nodes
   */
  getDistance(e, t, r) {
    const i = [...this.getPathNodes(e, t), this.get(e, t)], s = [...this.getPathNodes(e, r), this.get(e, r)], n = [...i].reverse().find((d) => s.includes(d));
    if (!n)
      return -1;
    const h = i.findIndex((d) => d.id === n.id), l = s.findIndex((d) => d.id === n.id);
    return i.length - h - 1 + (s.length - l - 1);
  }
  /**
   * Helper method to deep clone object
   * @param obj - object to be cloned
   * @private
   * @returns deep cloned object
   */
  static deepCopy(e) {
    return JSON.parse(JSON.stringify(e));
  }
}
export {
  o as TreeUtils
};
